/** 
 * 
 * Copyright 2004 Protique Ltd
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License. 
 * 
 **/
package org.codehaus.activemq.util;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.net.InetAddress;
import java.net.ServerSocket;

/**
 * Generator for globally unique Strings
 *
 * @version $Revision: 1.2 $
 */
public class IdGenerator {
    private static final Log log = LogFactory.getLog(IdGenerator.class);
    private static final String UNIQUE_STUB;
    private static int instanceCount;
    private static String hostName;
    private String seed;
    private long sequence;

    static {
        String stub = "";
        boolean canAccessSystemProps = true;
        try {
            SecurityManager sm = System.getSecurityManager();
            if (sm != null) {
                sm.checkPropertiesAccess();
            }
        }
        catch (SecurityException se) {
            canAccessSystemProps = false;
        }
        if (canAccessSystemProps) {
            try {
                hostName = InetAddress.getLocalHost().getHostName();
                ServerSocket ss = new ServerSocket(0);
                stub = "ID:" + hostName + "-" + ss.getLocalPort() + "-" + System.currentTimeMillis() + "-";
                Thread.sleep(100);
                ss.close();
            }
            catch (Exception ioe) {
                log.warn("could not generate unique stub", ioe);
            }
        }
        else {
            hostName = "localhost";
            stub = "ID:" + hostName + "-1-" + System.currentTimeMillis() + "-";
        }
        UNIQUE_STUB = stub;
    }

    /**
     * As we have to find the hostname as a side-affect of generating a unique stub, we allow it's easy retrevial here
     *
     * @return the local host name
     */
    public static String getHostName() {
        return hostName;
    }

    /**
     * Construct an IdGenerator
     */
    public IdGenerator() {
        synchronized (UNIQUE_STUB) {
            this.seed = UNIQUE_STUB + (instanceCount++) + ":";
        }
    }
    
    /**
     * Construct an IdGenerator using the seed provided
     * @param seed
     */
    public IdGenerator(String seed){
        this.seed = seed;
    }
    

    /**
     * Generate a unqiue id
     *
     * @return a unique id
     */
    public synchronized String generateId() {
        return this.seed + (this.sequence++);
    }

    /**
     * @return the unique seed used by this generator
     */
    public String getSeed() {
        return seed;
    }
    
    /**
     * @return the next sequence
     */
    public synchronized long getNextSequence(){
        return this.sequence++;
    }

    /**
     * From a generated id - return the seed (i.e. minus the count)
     *
     * @param id the generated identifer
     * @return
     */
    public static String getSeedFromId(String id) {
        String result = id;
        if (id != null) {
            int index = id.lastIndexOf(':');
            if (index > 0 && (index + 1) < id.length()) {
                result = id.substring(0, index + 1);
            }
        }
        return result;
    }

    /**
     * From a generated id - return the generator count
     *
     * @param id
     * @return the count
     */
    public static long getCountFromId(String id) {
        long result = -1;
        if (id != null) {
            int index = id.lastIndexOf(':');

            if (index > 0 && (index + 1) < id.length()) {
                String numStr = id.substring(index + 1, id.length());
                result = Long.parseLong(numStr);
            }
        }
        return result;
    }

    /**
     * Does a proper compare on the ids
     *
     * @param id1
     * @param id2
     * @return
     */

    public static int compare(String id1, String id2) {
        int result = -1;
        String seed1 = IdGenerator.getSeedFromId(id1);
        String seed2 = IdGenerator.getSeedFromId(id2);
        if (seed1 != null && seed2 != null) {
            result = seed1.compareTo(seed2);
            if (result == 0) {
                long count1 = IdGenerator.getCountFromId(id1);
                long count2 = IdGenerator.getCountFromId(id2);
                result = (int) (count1 - count2);
            }
        }
        return result;

    }
}
