/**
 *
 * Copyright 2004 Hiram Chirino
 * Copyright 2004 Protique Ltd
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 **/
package org.codehaus.activemq.transport.gnet;

import org.apache.geronimo.network.SelectorManager;
import org.apache.geronimo.pool.ClockPool;
import org.apache.geronimo.pool.ThreadPool;
import org.codehaus.activemq.io.WireFormat;
import org.codehaus.activemq.transport.TransportChannel;
import org.codehaus.activemq.transport.TransportChannelFactorySupport;

import javax.jms.JMSException;
import java.net.URI;

/**
 * An implementation of a TransportChannelFactory which uses
 * the Geronimo network layer for connectivity.
 *
 * @version $Revision: 1.2 $
 */
public class GTransportChannelFactory extends TransportChannelFactorySupport {

    private static ThreadPool threadPool;
    private static ClockPool clockPool;
    private static SelectorManager selectorManager;

    public static void init(SelectorManager sm, ThreadPool tp,
                            ClockPool cp) throws IllegalArgumentException {

        if (sm == null || tp == null || cp == null) {
            throw new IllegalArgumentException();
        }

        selectorManager = sm;
        threadPool = tp;
        clockPool = cp;
    }

    private static void init() throws Exception {

        // Don't init if allready done.
        if (threadPool != null) {
            return;
        }

        threadPool = new ThreadPool();
        threadPool.setKeepAliveTime(1 * 1000);
        threadPool.setPoolSize(5);
        threadPool.setPoolName("C Pool");

        clockPool = new ClockPool();
        clockPool.setPoolName("C Clock");

        selectorManager = new SelectorManager();
        selectorManager.setThreadPool(threadPool);
        selectorManager.setThreadName("C Manager");
        selectorManager.setTimeout(1000);

        threadPool.doStart();
        clockPool.doStart();
        selectorManager.doStart();

    }

    /**
     * Create a Channel to a remote Node - e.g. a Broker
     *
     * @param wireFormat
     * @param remoteLocation
     * @return the TransportChannel bound to the remote node
     * @throws JMSException
     */
    public TransportChannel create(WireFormat wireFormat, URI remoteLocation) throws JMSException {
        return create(null, remoteLocation);
    }

    /**
     * Create a Channel to a remote Node - e.g. a Broker
     *
     * @param wireFormat
     * @param remoteLocation
     * @param localLocation  -
     *                       e.g. local InetAddress and local port
     * @return the TransportChannel bound to the remote node
     * @throws JMSException
     */
    public TransportChannel create(WireFormat wireFormat, URI remoteLocation, URI localLocation) throws JMSException {
        try {
            init();
            return populateProperties(new GTransportChannel(wireFormat, remoteLocation, localLocation, selectorManager, threadPool, clockPool), remoteLocation);
        }
        catch (Exception e) {
            JMSException ex = new JMSException(e.getMessage());
            ex.setLinkedException(e);
            throw ex;
        }
    }

    public boolean requiresEmbeddedBroker() {
        return false;
    }
}